#include "sqliteInt.h"

static void sqliteTriggerAddVar(Trigger *pTrig, Var *pVar)
{
	/* Link in list */
	if (0 == pTrig->pVars)
		pTrig->pVars = pVar;
	else {
		pVar->pNext = pTrig->pVars;
		pTrig->pVars = pVar;
	}
}

static Var* sqliteVarNew(Parse *pParse, const char* zName, int iLen) 
{
	Var *pVar;

	assert(pParse!=0);
	assert(zName!=0);
	assert(iLen>=0);

	/* Variable name length <= VAR_NAME_LEN */
	if (iLen > VAR_NAME_LEN){
		pParse->rc = SQLITE_ERROR;
		goto error;
	}

	pVar = (Var*) sqliteMalloc(sizeof(Var));
	if (0==pVar) {
		pParse->rc = SQLITE_NOMEM;
		goto error;
	}

	memcpy (pVar->zName, zName, iLen);
	pVar->n = iLen;
	pVar->state = VC_UNDECLARED;

	return pVar;

error:
	if (pParse->rc>0) {
		Trigger *pTrig = sqlite3TriggerFromParse(pParse);
		assert(pTrig!=0);
		if (SQLITE_ERROR==pParse->rc){
			sqlite3ErrorMsg(pParse, "trigger: %T(%d): variable name length > %d", 
											&pTrig->nameToken, pParse->nLine, VAR_NAME_LEN);
		}
		if (SQLITE_NOMEM==pParse->rc){
			sqlite3ErrorMsg(pParse, "trigger: %T(%d): sqliteVarNew: not enough memory",
											&pTrig->nameToken, pParse->nLine);
		}
	}
	return 0;
}

void sqlite3VarDelete(Var *pVar)
{
	if (pVar!=0){
		if (pVar->pExpr!=0) {
			sqlite3ExprDelete(pVar->pExpr);
			pVar->pExpr = 0;
		}
		sqlite3SelectDelete(pVar->pSelect);
		sqliteFree(pVar);
	}
}

void sqlite3DeleteVars(Trigger *pTrig)
{
	Var *pVar;
	
	assert(pTrig!=0);

	pVar = pTrig->pVars;
	while (pVar) {
		pTrig->pVars = pVar->pNext;
		sqlite3VarDelete(pVar);
		pVar = pTrig->pVars;
	}
}

Var* sqlite3TrigVarFindByToken(Trigger *pTrig, Token *pTkVarName)
{
	Var *pVar;

	assert(pTrig!=0);
	assert(pTkVarName!=0);

	for(pVar = pTrig->pVars; pVar!=0; pVar = pVar->pNext) {
		int i = (pTkVarName->n > pVar->n) ? pTkVarName->n : pVar->n;
		if (sqlite3StrNICmp(pVar->zName, pTkVarName->z, i)==0)
			return pVar;
	}

	return 0;
}

Var* sqlite3VarFindByStr(Parse *pParse, const char *zName, int iLen)
{
	Trigger *pTrig;
	Token token;

	assert(pParse!=0);
	assert(zName!=0);
	assert(iLen>0);

	pTrig = sqlite3TriggerFromParse(pParse);
	if (0==pTrig)
		return 0;

	token.z = zName;
	token.n = iLen;
	
	return sqlite3TrigVarFindByToken(pTrig, &token);
}

Var* sqlite3VarFindByToken(Parse *pParse, Token *pTkVarName)
{
	Trigger *pTrig;

	assert(pParse!=0);
	assert(pTkVarName!=0);

	pTrig = sqlite3TriggerFromParse(pParse);
	if (0==pTrig)
		return 0;

	return sqlite3TrigVarFindByToken(pTrig, pTkVarName);
}


Var* sqlite3VarAssignExpr(Parse *pParse, Token *pTkVarName, Expr *pExpr)
{
	Var* pDeclVar;
	
	assert(pParse!=0);
	assert(pTkVarName!=0);

	if (pParse->rc!=SQLITE_OK || pParse->nErr!=0 || 0==pParse->pNewTrigger)
		return 0;

	if (0==pExpr)
		return 0;

	pDeclVar = sqlite3VarFindByToken (pParse, pTkVarName);
	if (0==pDeclVar){
		sqlite3ErrorMsg(pParse, "trigger: %T(%d): variable %T not declared", 
										&pParse->pNewTrigger->nameToken, pParse->nLine, pTkVarName);
		pParse->rc = SQLITE_ERROR;
		return 0;
	}

	if (pDeclVar->state & C_DECLARED) {
		sqlite3ErrorMsg(pParse, "trigger: %T(%d): redefinition: %T declared as cursor", 
										&pParse->pNewTrigger->nameToken, pParse->nLine, pTkVarName);
		pParse->rc = SQLITE_ERROR;
		return 0;
	}

	if (sqlite3StrNICmp("@@count", pTkVarName->z, pTkVarName->n)==0) {
		sqlite3ErrorMsg(pParse, "trigger %T(%d): can not assign system variable @@count", 
										&pParse->pNewTrigger->nameToken, pParse->nLine, pTkVarName);
		pParse->rc = SQLITE_ERROR;
		return 0;
	}

	sqlite3ExprDelete (pDeclVar->pExpr);	/* We save pExpr in TriggerVarAssignStep */
	pDeclVar->pExpr = pExpr;
	pDeclVar->state = (V_DECLARED | V_ASSIGNED);

	if (0==pDeclVar->i)
		pDeclVar->i = ++pParse->nVar;
	
	if (0==pExpr->iTable)
		pExpr->iTable = pDeclVar->i;
	
	if ( pDeclVar->i > pParse->nVar ){
		pParse->nVar = pDeclVar->i;
	}

	return pDeclVar;
}

Var* sqlite3VarDup(Parse *pParse, Var *pVar)
{
	Var *pNew;

	assert(pParse!=0);
	assert(pVar!=0);

	pNew = sqliteVarNew(pParse, pVar->zName, pVar->n);
	if (0==pNew) 
		return 0;

	pNew->pExpr		= sqlite3ExprDup (pVar->pExpr);
	pNew->pSelect = sqlite3SelectDup (pVar->pSelect);
	pNew->i				= pVar->i;
	pNew->state		= pVar->state;

	return pNew;
}

Var* sqlite3VarAddFromToken (Parse *pParse, Token *pTkVarName)
{
	Trigger *pTrig;
	Var *pNewVar;


	assert(pParse!=0);
	assert(pTkVarName!=0);

	pNewVar = sqliteVarNew(pParse, pTkVarName->z, pTkVarName->n);
	if (pNewVar==0)
		return 0;

	pTrig = sqlite3TriggerFromParse(pParse);
	if (0==pTrig)
		return 0;

	sqliteTriggerAddVar(pTrig, pNewVar);

	return pNewVar;
}

Var* sqlite3VarAddFromStr (Parse *pParse, const char *p, int i)
{
	Trigger *pTrig;
	Var *pNewVar;

	assert(pParse!=0);
	assert(pParse->pNewTrigger!=0);
	assert(p!=0);
	assert(i!=0);

	pNewVar = sqliteVarNew(pParse,p,i);
	if (0==pNewVar)
		return 0;

	pTrig = sqlite3TriggerFromParse(pParse);
	if (0==pTrig)
		return 0;

	sqliteTriggerAddVar(pTrig, pNewVar);

	return pNewVar;
}

/*
 *	Return token length with spaces after it.
 *  ex: insert into t1 values( @CursorName.FieldName   , @a.aaa   )
 *                             ^^^^^^^^^^^^^^^^^^^^^^^^  ^^^^^^^^^
 *      declare @Var     ;
 *							^^^^^^^^^
 */
int sqlite3TokenLenWithSpace(Token *pToken) {
	int i=pToken->n;
	const unsigned char *pStr = pToken->z + i;
	
	if (pToken->dyn==1) return i;

	for (;*pStr==' '; i++){pStr++;}
	return i;
}
